#include <argz.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <libintl.h>
#include <limits.h>
#include <locale.h>
#include <regex.h>
#include <signal.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <unistd.h>

#ifndef DIET //Not dietlibc.
#include <obstack.h>
#endif

//Include postgresql libs for database operations.
#include <pgsql/libpq-fe.h>

#define PG_ERR      1
#ifndef PG_ERR //pgsql has a lib problem.
#include <pgsql/server/catalog/pg_type.h>
#endif

//Include libhx for various functions.
#include <libHX-3.15/libHX.h>
#include <libHX-3.15/libHX/string.h>
#include <libHX-3.15/libHX/io.h>

//Include magic for file types.
#include <magic.h>


//My includes.
#include "main.h"
#include "scanner.h"


#ifndef ALLH //Only allow inclusion once.
#define ALLH
/*
 * Now to define some error messages.
 * Remember that 0x3 brings positive results in the range 1 through 3
 * therefore it becomes necessary to use an equal 3, like this:
 * if((x & 0x3) == 3)
 * 0x2 needs to be like this:
 * if(x & 0x3)
 * and 0x1 is an exeption to this rule.
 */

/*
 * Deciphering w3dd messages:
 * 
 * The syntax for most w3dd messages is:
 * 
 * ________________________________________
 * | item  | class | message abbreviation |
 * |_______|_______|______________________|
 * | chars | 1-2   | 2-4                  |
 * |_______|_______|______________________|
 * 
 * Where class is one of:
 * 
 * __________________________________________________________________________
 * | Abbr | acronym for |   Meaning of                  | Terminate program |
 * |______|_____________|_______________________________|___________________|
 * |      |             | A Universal message that can  |                   |
 * |      |             | be printed for all instances. |   Maybe.          |
 * |______|_____________|_______________________________|___________________|
 * |      |             |   Debug information           |                   |
 * | DB   | DEBUG       |   (see additional notes.)     |   Never.          |
 * |______|_____________|_______________________________|___________________|
 * |      |             |   Debug information           |                   |
 * | D    | DEBUG       |   (see additional notes.)     |   Never.          |
 * |______|_____________|_______________________________|___________________|
 * |      |             |   Error, generally a user     |                   |
 * | E    | ERROR       |   created bad, incorrect or   |                   |
 * |      |             |   missing data condition.     |   Maybe.          |
 * |______|_____________|_______________________________|___________________|
 * |      |             |   Fatal, unrecoverable/not    |                   |
 * | F    | FATAL       |   ever supposed to happen     |   Always.         |
 * |      |             |   condition.                  |                   |
 * |______|_____________|_______________________________|___________________|
 * |      |  FATAL      |   Fatal, unrecoverable/not    |                   |
 * | FP   |  POSTGRESQL |   ever supposed to happen     |   Always.         |
 * |      |  or PIPE    |   condition.                  |                   |
 * |______|_____________|_______________________________|___________________|
 * |      |             |   Verbose, talkative, what    |                   |
 * | V    | VERBOSE     |   w3dd is up to kinda message.|   Never.          |
 * |______|_____________|_______________________________|___________________|
 * |      |             |   Warning, not supposed       |                   |
 * |      |             |   to occur but can be         |                   |
 * | W    | WARNING     |   corrected, generally caused |   Never.          |
 * |      |             |   by bad data from the user or|                   |
 * |      |             |   another source.             |                   |
 * |______|_____________|_______________________________|___________________|
 * 
 * Debug info can be between one and three where 1 is just a little and three
 * dumps huge amounts of debugging information onto your terminal. Setting
 * Debug does not print verbose information. If you want both pass both.
 * 
 * 
 * And message abbreviation is anything that makes sense to yours truly.
 * 
 */


//Universals
#define END   "'.\n"
#define SPACE "' '"
#define ORR   "' or '"
#define FON   "' for option '"

#define ERRRE "Errno reports: %s\n"
#define RGERR "%s: Regerr reports: %s\n"
#define WHAT  "'%s': In valid responce '%c'. Try again.\n\n"
#define LMERR "Libmagic reports: %s\n"
#define PQEMR "PQerrorMessage reports: %s\n"

//Debugs
#define DBGE  "' == '"
#define DBGB  ": DEBUG: big == '"
#define DBGL  ": DEBUG: element == '"
#define DBGN  ": DEBUG: next == '"
#define DBGO  ": DEBUG: '"
#define DBGS  ": DEBUG: small == '"
#define DBGR  ": DEBUG: root == '"
#define DBGT  ": DEBUG: string == '"
#define DBGU  ": DEBUG: number == '"
#define DBSE  ": DEBUG: size == '"
#define DBDE  ": DEBUG: domain == '"

#define DBGER "%s: DEBUG: EINTR file '%s' line '%i'\n"
#define DBGTA "%s: DEBUG: Option == '%s' in file '%s' with line number '%i'\n"
#define DBGGD "%s: DEBUG: Getopt finished results are '"
#define DBGDC "%s: DEBUG: Dumping complete config from memory.\n"
#define DBEIM "%s: DEBUG: If it has an extention the " \
"extention is missing from the string: %s\n"
#define DBEIA "%s: DEBUG: Extention '%s' allowed.\n",
#define DBEID "%s: DEBUG: Extention '%s' disallowed.\n"
#define DBENA "%s: DEBUG: Extention '%s' is not up for asking as " \
"it is in domain '%s' so, it is being fetched.\n"
#define DBRNM "%s: DEBUG: String '%s' did not match regulare expression.\n"
#define DBGRM "%s: DEBUG: String '%s' matched regulare expression.\n"
#define DBGSS "%s: DEBUG: String '%s' has been split into %i parts.\n"


//Verbose
#define VOFAI "%s: Verbose: Argument for option '%s' is '%s'\n"
#define VCCC  "%s: Verbose: Checking curl compatibility.\n"
#define VSYSC "%s: Verbose: Sysconfig defines the maximum values for " \
"argv and children as %hi %li\n"
#define VFS   "%s: Verbose: Forking succeeded, this is the parent.\n"
#define VCCCC "%s: Verbose: Curl compatibility check completed.\n"
#define VBHTM "%s: Verbose: Missing tag(s) in the html segment: \n\n'%s'\n\n"\
"consider looking into it.\n"
#define VEI   "%s: Verbose: File extension is '%s'\n"
#define VSNMS "%s: Verbose: String '%s' did not match any of the strings " \
"given for option '%s'.\n"
#define VSNMR "%s: Verbose: String '%s' did not match any of the regular " \
"expressions given for option '%s'.\n"
#define VTDII "%s: Verbose: The default is to ignore.\n"
#define VTDIF "%s: Verbose: The default is to fetch.\n"


//Warnings
#define WCSNR "%s: Warning: Control should not reach this statement. "\
"File '%s' line '%i'\n"
#define WNPRT "%s: Warning: No port given: assuming 1080 and if that fails " \
"then 8080.\n"
#define WNPGA "%s: Warning: No protocol given: assuming 'http://'.\n"
#define WNET  "%s: Warning: Not enough threads allowed. Lowering used amount.\n"
#define WACFS "%s: Warning: Allowed core file size is '%i'.\n Consider "\
"setting it to a higher value with\n\n\t$ulimit -c NUMB\n\n"
#define WCVNR "%s: Warning: Curl version is not recent enough though it may "\
"run correctly depending on the configuration that you have selected.\n"\
"If you find that this is an error on my programs part please report it.\n"\
"I recommend at least 7.28.1 but got ''\n"
#define WTP   "%s: Warning: The feature '%s' is not supported. "\
"If you don't need it then you can safely ignore this message other "\
"wise you need to recompile curl with support for it.\n"
#define WOIAO "%s: Waring: Overflow in the addition of %lli and %lli.\n"
#define WLHDV "%s: Waring: '%s' has a different version then the " \
"version it was compiled against.\nThe version compiled against was " \
"'%i' the version that your system is using is '%i'.\nThis may " \
"or may not cause problems.\n"

//Fatals
#define WRONG "%s: Fatal: This statement should not ever be reached. "\
"File '%s' Line '%i'\n"
#define FDNE  "%s: Fatal: The file '%s' does not exist.\n"
#define FCAM  "%s: Fatal: Cannot allocate memory.\n"
#define FSFER "%s: Fatal: Lseek failed: Errno reports: '%s'\n"
#define FTUPV "%s: Fatal: Unpermitted value for the variable type '%i'\n"
#define FDDF  "%s: Fatal: Duplicating descriptor failed.\n"
#define FEFER "%s: Fatal: Execlp failed.\n"
#define FFFER "%s: Fatal: Forking failed.\n"
#define FUKCS "%s: Fatal: Unprocessable curl version string '%s'.\n"\
"Expected a value of '%s'\n"
#define FPF   "%s: Fatal: Pipe failed.\n"
#define FCCD  "%s: Fatal: Can't close descriptor.\n"
#define FFPD  "%s: Fatal: File '%s' can not be opened.\n"
#define FCRFF "%s: Fatal: File '%s' can not be read from.\n"
#define FCWTF "%s: Fatal: File '%s' can not be written to.\n"
#define FCSIF "%s: Fatal: File '%s' can not be seeked in.\n"
#define FCOD  "%s: Fatal: Directory '%s' can not be opened.\n"
#define FCOD2 "%s: Fatal: File '%s' can not be dup2-ed with stderr.\n"
#define FCOD1 "%s: Fatal: File '%s' can not be dup2-ed with stdout.\n"
#define FWF   "%s: Fatal: Wait failed.\n"
#define FCPR  "%s: Fatal: Child process '%s' returned '%i'.\n"
#define FCCR  "%s: Fatal: curl command '%s' returned %i.\n"
#define FSROM "%s: Fatal: String that ran regexec out of memory is '%s'."
#define FSBF  "%s: Fatal: setvbuf failed errno reports: %s\n"
#define FILF  "%s: Fatal: Initialization of %s failed.\n"
#define FCCPG "%s: Fatal: Cannot connect to postgresql server.\n"
#define FPCCT "%s: Fatal: Cannot create table, command '%s'.\n"
#define FPCDT "%s: Fatal: Cannot drop table, command '%s'.\n"
#define FPCPS "%s: Fatal: Cannot prepare statement, command '%s'.\n"
#define FPCID "%s: Fatal: Cannot insert data into table.\n"

//Errors
#define ENAFO "%s: Error: No argument for option '%s'\n"
#define ENNFO "%s: Error: No positive integer argument for option '%s'\n"
#define EAFO  "%s: Error: Argument for option '%s' should be "\
"between '%lli' and '%lli'\n"
#define EATS  "%s: Error: Arguments '%s' and '%s' are the "\
"same for option '%s'\n"
#define EARGL "%s: Error: Argument '%lli' too long for option '%s\n"
#define EARGS "%s: Error: Argument '%lli' too short for option '%s'\n"
#define ENSFO "%s: Error: No string for option '%s'\n"
#define ENREF "%s: Error: No regular expression for argument '%s'\n"
#define EOAG  "%s: Error: Option '%s' has already been given "\
"and may not be given again.\n"
#define EUAFO "%s: Error: Unknown argument '%s' for option '%s'\n"
#define EUO   "%s: Error: Unknown option '%s'\n"
#define ENSFU "%s: Error: No source for urls specified.\n"
#define ENYNA "%s: Error: No yes/no argument for option '%s'\n"
#define EATL  "%s: Error: Argument '%s' is to long a string for option '%s'\n"
#define ENCFO "%s: Error: No case specified for option '%s'\n"
#define ESTL  "%s: Error: File '%s' has a string '%s' that's way to long.\n"
#define EOAII "%s: Error: Option's '%s' argument '%s' is incomplete.\n"
#define ECCRE "%s: Error: Cannot compile regular expression '%s'\n"
#define EIVFO "%s: Error: Invalid argument '%s' for option '%s'\n"
#define ESS   "%s: Error: String '%s' was split into %i strings instead of 2.\n"
#define ECNAF "%s: Error: The char '%c' is not allowed in the file %s\n"
#define EAION "%s: Error: String argument '%s' is identical  to " \
"preexisting value but '%i' argument is not for option '%s'.\n"
#define ESTNS "%s: Error: simultaneous_transfers not set or zero.\n"
#define ERLNS "%s: Error: recursion_level not set or zero.\n"
#define ENLFS "%s: Error: No logfile set.\n"


#define STDIN  0
#define STDOUT 1
#define STDERR 2

//For printing version info.
#define VERSION(stra, strb) (stra) ## "0.0.1" ## (strb)


//For error reporting.
#define REGCOMP 1
#define REGEXEC 0

//If I define anything that is optional it goes here.
#define UCLIBC      1
#define DIET        1
#define HUGESTRING  1
#define REGEXSUB_ZERO 1

//end


//How much memory can I use on a string?
#ifdef HUGESTRING //Longest string. Normally 65535 bytes.
#define STRLENGHT USHRT_MAX
#elif LONGLONGSTRING //Long long strings.
#define STRLENGHT 16384
#elif  LONGSTRING //Long strings.
#define STRLENGHT 4096
#elif  NORMALSTRING //Normal strings.
#define STRLENGHT 1024
#elif  SHORTSTRING //Short strings. Normally 255 bytes.
#define STRLENGHT UCHAR_MAX
#else  //Nothing defined!
#error Define a string size appopriate for your computer. I recommend LONGSTRING.
#define STRLENGHT 1
#endif


//How much memory can I use on regex subexpression matching information?
#ifdef REGEXSUB_ZERO //No mtaching information.
#define REGMATCH_T_MAX 0
#elif  REGEXSUB_SMALL //Maybe too little,
#define REGMATCH_T_MAX UCHAR_MAX //but it works for those with no memory to spare.
#elif REGEXSUB_MODERATE //Should be enough.
#define REGMATCH_T_MAX 1024
#elif  REGEXSUB_LARGE //Who could want more?
#define REGMATCH_T_MAX 4096
#elif  REGEXSUB_HUGE //Way too much for any normal person.
#define REGMATCH_T_MAX 16384
#else  //Nothing defined!
#error Define a regex subexpression matching infomation size \
appopriate for your computer. I recommend ZERO, but you may want more. \
If you set the value too low, but greater then zero, it will cause more \
memory consuption, not less. But, If you do not include many subexpressions \
in your regular expressions this will not be a problem.
#endif

//Uclibc does not have the gettext functions and I'm not sure how to add them.
#ifdef UCLIBC //A libc.
#define _(str) str
#else //Gnu-libc or diet libc.
#define _(str) gettext(str)
#endif


/*
 * Checks fuction write_short_str2_des.
 * It depends on the existance of my_name.
 */

#define STR_ERRNO_CHECK() \
if(errno)\
{\
    GGGerrno_str_ptr = strerror(errno);\
    fprintf(stderr, _(&ERRRE), GGGerrno_str_ptr);\
    (void)raise(SIGABRT);\
}


/*
 * Prints an error message from an uncompilable regex.
 * It handles all errors internally.
 * !alloc
 * 
 * It calls: fprintf, strerror, and regerr. It is: MT-AS-AC-Unsafe.
 */
void regerr_reporter(const regex_t *regex_t_ptr, \
            unsigned char *str_that_caused_error, \
              int regcomp_return, const bool cause );

/*
 * Prints error condition on all files.
 * It handles all errors internally.
 * !alloc
 * 
 * It exits in case of an unrecoverable error.
 * 
 * It calls: fprintf and strerror. It is: MT-AS-AC-Unsafe.
 */
void check_errnof(const unsigned char * const message, \
                  const unsigned char * const file_name );

#define obstack_chunk_alloc malloc
#define obstack_chunk_free  free


/*
 * Check the strings length to avoid overflow.
 * It depends on my_name. The variable length must be length
 * of the array minus one.
 */

#define CHCKSTR(tmp_var_i, tmparg, length) \
if((tmp_var_i) == (length)) \
{ \
    (tmparg)[(tmp_var_i)] = '\0'; \
    fprintf(stderr, _(&EATL), my_name, tmparg, option);\
    exit(EXIT_FAILURE); \
}



/*
 * These are some smaller structs to supliment the larger one.
 * 
 * Legend
 * 
 * sl singly linked
 * tl tree   linked
 * nl not    linked
 */


//This one holds the passwords.
struct tri_str_tl
{
    unsigned char *domain;
    unsigned char *user;
    unsigned char *password;
    struct tri_str_tl *big;
    struct tri_str_tl *small;
};


union numbs
{
    signed short int sint;
    unsigned short int usint;
    unsigned long int ulint;
    unsigned long long int ullint;
};


//This one holds a ptr to string and a number.
struct domain_numb_tl
{
     unsigned char *domain; //Normally holds a domain string.
     union numbs anumber;
     struct domain_numb_tl *big;
     struct domain_numb_tl *small;
};

struct domain_element_str_numb_tl
{
    unsigned char *domain;
    unsigned char *element;
    unsigned char *str;
    union numbs anumber;
    struct domain_element_str_numb_tl *big;
    struct domain_element_str_numb_tl *small;
};

struct domain_sl
{
    unsigned char *domain;
    struct domain_sl *next;
};

struct domain_array_ext_numb_tl
{
    unsigned char *ext;
    union numbs size;
    struct domain_sl *the_domains[255];
    struct domain_array_ext_numb_tl *big;
    struct domain_array_ext_numb_tl *small;
};


/*
 * Research:
 * 
 * #include <stdio.h>
 * #include <regex.h>
 * 
 * int main(void)
 * {
 *      regex_t *ptr;
 *      printf("reg %i ptr %i\n", (int)sizeof(regex_t), (int)sizeof(ptr));
 *      return(0);
 * }
 * 
 * Results:
 * 
 * reg 64 ptr 8
 * 
 * Suggestion:
 * Reduce The amount of regex_t struct repetitions.
 * Unfortunately, I've not found a good way to implement this.
 * Example:
 * 
 * Legend:
 * 
 * = Array  horizontal
 * | sl list vertical
 * - ptr horizontal
 * / tl list diagonal
 * \ tl list diagonal
 * D Domain regex_domain_tl
 * A array containing pointers to type: regexp_sl
 * S Singly linked list of regexps: regexp_sl
 * R Regular expression.
 * 
 * 
 * A3==========A2==========D1===========A1      
 * |           |          /  \          |       
 * |           |         /    \         |       
 * S3----R1----S2       /      \       S1-R3--S6
 * |         A1========D2=A2   D3===A1===A2===A3
 * |         |            |         |    |      
 * S4----R2--S5           |         |    |      
 *                        S9----R4--S7   S8---R5
 * 
 * 
 * Now before I go ascii art happy I'm going to quit.
 * 
 * I might have used a doubly linked list instaed of an array but that
 * would slow down access time on top of making for some rather confusing
 * code (though it would have saved some memory.)
 */

struct regexp_sl
{
    regex_t extended_reg;
    struct regexp_sl *next;
};


/*
 * All of them must have the same domain.
 * regexp_sl MUST be 257 because it starts from 0 and ends at 256.
 * It's kind of hard on the RAM though, taking up a total of 2080 bytes
 * on my system.
 */


struct regex_array_domain_tl
{
    unsigned char *domain;
    struct regexp_sl *the_regs[256];
    struct regex_array_domain_tl *big;
    struct regex_array_domain_tl *small;
};


//This is for those that don't need a depth.
struct regex_domain_tl
{
    unsigned char *domain;
    struct regexp_sl *root;
    struct regex_domain_tl *big;
    struct regex_domain_tl *small;
};

//It will store the text/html tags in addition to domain names.
struct domain_tl
{
    unsigned char *domain;
    struct domain_tl *big;
    struct domain_tl *small;
};

struct string_domain_tl
{
    unsigned char *str;
    struct domain_tl *root;
    struct string_domain_tl *big;
    struct string_domain_tl *small;
};


/*
 * The main configuration reteived from the configuration files is
 * stored in this struct. It's slightly larger than a kilobyte in size.
 */

struct w3dd_config
{
    unsigned short int simultaneous_transfers;
    unsigned short int recursion_level;
    unsigned short int recursion_level_external;
    unsigned short int recursion_level_horizontal;
    unsigned short int recursion_level_question_mark;
    unsigned short int recursion_level_horizontal_external;
    unsigned short int recursion_level_question_mark_external;
    unsigned short int four_04_retries;
    unsigned short int width;
    unsigned short int height;
    unsigned short int resolution;
    unsigned long int colors;
    unsigned long int keep_alive_optimization_amount;
    
    unsigned long int force_total_time_limit_to;
    
    unsigned long int argv_max;
    
    /*
     * On 64 bit platforms the pointers are 64 bits long but on 32 bit
     * platforms the pointers are 32 bits long. I'm putting them here
     * because if they are 64 bits long then putting the
     * unsigned long long ints after them does no harm and if they are
     * 32 bits then the should go right before or right after the
     * unsigned long ints. So, in either case there should be no padding
     * necesary.
     */
    
    //Proxy stuff.
    unsigned char *protocol;
    unsigned char *proxyhost;
    unsigned char *port;
    unsigned char *user;
    unsigned char *password;
    
    unsigned char *random_source;
    unsigned char *path_to_download_to;
    unsigned char *logfile;
    unsigned char *filter_file;
    unsigned char *advanced_options_file;
    unsigned char *cookies;
    unsigned char *max_redirects;
    unsigned char *login_page;
    unsigned char *update_type;
    unsigned char *path_to_curl;
    unsigned char *browser_defaults_to;
    unsigned char *bittorrent_cmd;
    unsigned char *retries;
    unsigned char *timeout;
    unsigned char *wait;
    unsigned char *retry_connrefused;
    unsigned char *wait_before_retry;
    unsigned char *limit_rate;
    unsigned char *limit_slow_downloads;
    
    unsigned char *exclude_file;
    unsigned char *include_file;
    unsigned char *allowed_hosts_to_span_file;
    unsigned char *disallowed_hosts_to_span_file;
    
    unsigned char *aspect_ratio;
    unsigned char *send_curl_these_additional_options;
    unsigned char *before_each_download_execute_this;
    unsigned char *after_each_download_execute_this;
    unsigned char *at_final_program_exit_execute_this;
    unsigned char *force_proxy_socks5_gssapi_service_name;
    unsigned char *ssl_ssh_passphrase;
    
    //More proxy stuff.
    unsigned char *force_http_proxy_1_0_address;
    unsigned char *force_http_proxy_1_0_port;
    
    //Config file's structs.
    struct tri_str_tl *httppass;
    struct tri_str_tl *ftppass;
    
    struct domain_numb_tl *recursion_level_domain;
    struct domain_numb_tl *recursion_level_horizontal_domain;
    struct domain_numb_tl *recursion_level_question_mark_domain;
    struct domain_numb_tl *keep_alive_optimization_amount_type; 
    struct domain_numb_tl *keep_alive_optimization_amount_extention;
    
    //Filter file's structs.
    struct domain_array_ext_numb_tl *get_larger_then;
    struct domain_array_ext_numb_tl *get_smaller_then;
    
    struct domain_element_str_numb_tl *discard_url_for_element_with_value;
    struct domain_element_str_numb_tl *discard_urls_for_element_with_value;
    
    struct domain_numb_tl *permit_host_path_ascension_limit;
    struct domain_numb_tl *ignore_lines_from_begining_to;
    struct domain_numb_tl *ignore_lines_from_end_to;
    
    struct domain_tl *permit_host_path_ascension_ask;
    struct domain_tl *if_name_is_same_get_image;
    struct domain_tl *if_name_is_same_get_video;
    struct domain_tl *if_name_is_same_get_audio;
    struct domain_tl *if_name_is_same_get_doc;
    struct domain_tl *dirty_site_page;
    struct domain_tl *dirty_site_pages;
    struct domain_tl *parse_java_script;
    struct domain_tl *dirty_site_pages_java;
    struct domain_tl *parse_php_script;
    struct domain_tl *parse_perl_script;
    struct domain_tl *allowed_hosts_to_span;
    struct domain_tl *disallowed_hosts_to_span;
    
    struct regexp_sl *allowed_hosts_to_span_reg;
    struct regexp_sl *disallowed_hosts_to_span_reg;
    
    struct regex_array_domain_tl *include;
    struct regex_array_domain_tl *exclude;
    
    struct string_domain_tl *file_extention_fetch;
    struct string_domain_tl *file_extention_ignore;
    struct domain_tl *fetch_dot;
    struct domain_tl *ignore_dot;
    
    struct regex_domain_tl *ascension_allowed;
    struct regex_domain_tl *ascension_disallowed;
    struct regex_domain_tl *fetch_java_link;
    struct regex_domain_tl *ignore_java_link;
    
    struct domain_tl *fetch_metalinks;
    struct domain_tl *ignore_metalinks;
    struct domain_tl *ask_for_no_known_file_ext;
    struct domain_tl *ask_for_no_known_host;
    struct domain_tl *ask_for_tiny_files;
    struct domain_tl *ask_for_no_known_metalinks;
    struct domain_tl *follow_only_next;
    struct domain_tl *permit_host_path_ascension_ask_first;
    
    //Advanced file's enums.
    enum languages langs[12];
    enum viewing_device devices[12];
    
    //Advanced file's structs.
    struct tri_str_tl *tlspass;
    
    struct domain_tl *force_http_1_0;
    struct domain_tl *force_tls_version_1;
    struct domain_tl *force_no_compression;
    struct domain_tl *force_no_dirs;
    struct domain_tl *force_http_digest_auth;
    struct domain_tl *force_no_eprt_and_lprt;
    struct domain_tl *force_pasv_only;
    struct domain_tl *force_no_referer_info;
    struct domain_tl *force_name_password_sending_apon_redirection;
    struct domain_tl *force_name_password_sending_apon_redirection_ask;
    struct domain_tl *force_respect_to_rfc_2616_301;
    struct domain_tl *force_respect_to_rfc_2616_302;
    struct domain_tl *force_http_raw;
    struct domain_tl *force_ftp_server_to_use_encryption_or_die_trying;
    struct domain_tl *force_ssl_allow_security_flaw_beast;
    struct domain_tl *force_use_insecure_ssl_connections_and_transfers;
    
    struct domain_tl *hack_ftp_crlf;
    struct domain_tl *hack_pret_command_prior_to_pasv_and_epsv;
    struct domain_tl *hack_http_ignore_content_length;
    struct domain_tl *hack_use_gss_negotiate_authentication;
    struct domain_tl *hack_ssl_no_session_id_caching;
    struct domain_tl *hack_other_than_http_use_proxy_connect_request;
    struct domain_tl *hack_proxy_socks5_gssapi_negotiation;
    struct domain_tl *noproxy;
    struct domain_tl *ftp_try_using_encryption_first;
    struct domain_tl *ignore_failure_of_command;
    
    struct string_domain_tl *force_these_ciphers_to_be_used;
    struct string_domain_tl *kerberos_delegation;
    struct string_domain_tl *force_ssl_version;
    struct string_domain_tl *force_tftp_block_size;
    struct string_domain_tl *force_minimum_speed;
    struct string_domain_tl *limit_rate_to;
    struct string_domain_tl *force_max_file_size;
    
    struct string_domain_tl *force_ftp_alternative_authenticating_command;
    struct string_domain_tl *force_ftp_ftps_file_getting_method;
    struct string_domain_tl *force_ftp_ssl_ccc_mode;
    struct string_domain_tl *force_proxy_authentication_method;
    struct string_domain_tl *use_crypto_engine_for_cipher_operations;
    struct string_domain_tl *pass_md5_chksum_of_the_remote_hosts_pub_key;
    struct string_domain_tl *use_private_key;
    struct string_domain_tl *use_pub_key;
    struct string_domain_tl *kerberos_authentication_and_use;
    struct string_domain_tl *ftp_ftps_send_command_before_transfer;
    struct string_domain_tl *ftp_ftps_send_command_after_transfer;
    struct string_domain_tl *ftp_send_command_after_changing_dir;
    
    /*
     * We use an unsigned long long int to simulate several bools
     * so that we don't have to worry about padding.
     */
    
    unsigned long long int options0_64;
    unsigned long long int quota;
    unsigned long long int four_04_retry_wait;
} GGGthe_config;



//Config file's bool opts.
#define RANDOM_WAIT                                         0x1
#define ROBOTSTXT                                           0x2
#define PROXY                                               0x4                 
#define NO_CACHE                                            0x8                 
#define MIRROR_LOWER_FIRST                                  0x10                
#define ADD_DIRS                                            0x20                
#define EXTENDED_GET                                        0x40                
#define EXTENDED_GET_POPULAR                                0x80                
#define SPAN_HOSTS                                          0x100               
#define TRY_TO_DETERMINE_FILE_TYPE_MANUALLY                 0x200
#define IF_UNCERTAIN_ASK_USER                               0x400
#define PARSE_JAVA_SCRIPT                                   0x800
#define ASK_FOR_JAVA_SCRIPT                                 0x1000
#define PARSE_PDF_FILES                                     0x2000
#define PARSE_PS_FILES                                      0x4000
#define PARSE_XML_FILES                                     0x8000
#define DOWNLOAD_TORRENTS                                   0x10000
#define DOWNLOAD_METALINKS                                  0x20000
#define PREFER_HTTPS                                        0x40000
#define ONLY_GO_DOWN                                        0x80000
#define UNKNOWN                                             0x100000
#define COPY_LINK_SHRINK                                    0x200000
#define PREFER_NO_FRAMES                                    0x400000

//The non bools that can only be given once. * Special note.
#define PATH_TO_DOWNLOAD_TO                                 0x800000
#define LOGFILE                                             0x1000000
#define FILTER_FILE                                         0x2000000
#define ADVANCED_OPTIONS_FILE                               0x4000000
#define COOKIES                                             0x8000000
#define LOGIN_PAGE                                          0x10000000
#define DOWNLOAD_TYPE                                       0x20000000
#define BROWSER_DEFAULTS_TO                                 0x40000000
#define PATH_TO_CURL                                        0x80000000
#define BITTORRENT_CMD                                      0x100000000
#define PROXY_ADDRESS                                       0x200000000
#define PROXY_PASS                                          0x400000000
#define RECURSION_LEVEL                                     0x800000000
#define RECURSION_LEVEL_EXTERNAL                            0x1000000000
#define RECURSION_LEVEL_QUESTION_MARK                       0x2000000000
#define RETRIES                                             0x4000000000
#define TIMEOUT                                             0x8000000000
#define LIMIT_SLOW_DOWNLOADS                                0x10000000000
#define WAIT                                                0x20000000000
#define LIMIT_RATE                                          0x40000000000
#define WAIT_BEFORE_RETRY                                   0x80000000000
#define QUOTA                                               0x100000000000
#define RETRY_CONNREFUSED                                   0x200000000000
#define MAX_REDIRECTS                                       0x400000000000
#define RECURSION_LEVEL_HORIZONTAL                          0x800000000000
#define RECURSION_LEVEL_HORIZONTAL_EXTERNAL                 0x2
#define RECURSION_LEVEL_QUESTION_MARK_EXTERNEL              0x1000000000000
#define FOUR_04_RETRY_WAIT                                  0x2000000000000
#define FOUR_04_RETRIES                                     0x4000000000000
#define KEEP_ALIVE_OPTIMIZATION_AMOUNT                      0x8000000000000
#define RECURSION_LEVEL_DOMAIN                              0x80000000000000
#define RECURSION_LEVEL_HORIZONTAL_DOMAIN                   0x100000000000000
#define RECURSION_LEVEL_QUESTION_MARK_DOMAIN                0x200000000000000
#define KEEP_ALIVE_OPTIMIZATION_AMOUNT_TYPE                 0x400000000000000
#define KEEP_ALIVE_OPTIMIZATION_AMOUNT_EXTENTION            0x800000000000000
#define SIMULTANEOUS_TRANSFERS                              0x1000000000000000
#define RANDOM_SOURCE                                       0x2000000000000000
//To note that no port was given.
#define PORT_NOT_GIVEN                                      0x4000000000000000
#define PORT_NOT_GIVEN_1_0                                  0x8000000000000000
//This leaves no bits untouched.


//Filter file's bool opts.
#define COMPARE_NAMES_IGNORING_DIR                          0x800000
#define COMPARE_SIZE_IF_SAME_NAME                           0x1000000
#define COMPARE_CONTENT_IF_SAME_NAME_AND_SIZE               0x2000000
#define COMPARE_CONTENT_IF_SAME_NAME_AND_SIZE_ONLY_HTML     0x4000000
#define IF_NAME_THE_SAME_AND_EXTENTION_IS_IMAGE_SKIP_THUMB  0x8000000
#define COMPARE_CONTENT_IF_SAME_NAME                        0x10000000
#define COMPARE_CONTENT_IF_SAME_SIZE                        0x20000000
#define ASK_FOR_NO_KNOWN_HOST                               0x40000000
#define ASK_FOR_NO_KNOWN_JAVA_LINK                          0x80000000
#define ASK_IF_BOTH_MATCH                                   0x100000000
#define IGNORE_WORTHLESS_PICTURES                           0x200000000
#define DO_NOT_MIRROR_UP_FROM_HORIZONTAL_LINKS              0x400000000
#define FETCH_ONLY_LATEST_VERSION                           0x800000000
#define IGNORE_COMPRESSION_SUFFIX                           0x1000000000

//The non bools that can only be given once. * Special note.
#define EXCLUDE_FILE                                        0x2000000000
#define INCLUDE_FILE                                        0x4000000000
#define ALLOWED_HOSTS_TO_SPAN_FILE                          0x8000000000
#define DISALLOWED_HOSTS_TO_SPAN_FILE                       0x10000000000
#define FORCE_THESE_CIPHERS_TO_BE_USED                      0x20000000000

//Advanced file's bool opts.
#define FORCE_HTTP_PROXY_NTLM                               0x2000000000
#define FORCE_NO_KEEP_ALIVE                                 0x4000000000
#define HACK_PROXY_SOCKS5_GSSAPI_NEGOTIATION                0x8000000000
#define ASK_FOR_NEW_CSS_TYPES                               0x10000000000
#define DOWNLOAD_AND_VERIFY_GPG_SIGN                        0x20000000000
#define DOWNLOAD_AND_VERIFY_SHA_SIGN                        0x40000000000
#define DOWNLOAD_AND_VERIFY_MD5_SIGN                        0x80000000000

//The non bools that can only be given once. * Special note.
#define ASPECT_RATIO                                        0x1
#define HEIGHT                                              0x2
#define WIDTH                                               0x4
#define RESOLUTION                                          0x8
#define COLORS                                              0x10
#define FORCE_PROXY_SOCKS5_GSSAPI_SERVICE_NAME              0x20
#define SSL_SSH_PASSPHRASE                                  0x40
#define SEND_CURL_THESE_ADDITIONAL_OPTIONS                  0x80
#define BEFORE_EACH_DOWNLOAD_EXECUTE_THIS                   0x100
#define AFTER_EACH_DOWNLOAD_EXECUTE_THIS                    0x200
#define AT_FINAL_PROGRAM_EXIT_EXECUTE_THIS                  0x400
#define FORCE_HTTP_PROXY_1_0                                0x800
#define FORCE_TOTAL_TIME_LIMIT_TO                           0x1000
#define VIEWING_DEVICE                                      0x2000
#define ACCEPT_LANGUAGE                                     0x4000


/*Binary values for download_type
#define START                                               0x8000000000
#define MERGE                                               0x10000000000
#define APPEND                                              0x20000000000
#define CONTINUE                                            0x40000000000
#define UPDATE                                              0x80000000000
#define REGET_ALL_HTML                                      0x100000000000
#define TIMESTAMPS                                          0x200000000000
#define SIZE                                                0x400000000000
*/

#endif


/* 
 * * Special note. The values under this heading need only be
 * different from the ones on the heading directly above it.
 * They are transitory values.
 * #define * The above influences the below but not visa versa.
#define */


